/*
 *  Copyright (c) 1990 Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  bind.c
 */

#ifndef lint 
static char copyright[] = "@(#) Copyright (c) 1990 Regents of the University of Michigan.\nAll rights reserved.\n";
#endif

#include "lber.h"
#include "ldap.h"
#include <stdio.h>
#include <string.h>
#ifdef MACOS
#include <stdlib.h>
#include "macos.h"
#else /* MACOS */
#ifdef DOS
#include "msdos.h"
#ifdef NCSA
#include "externs.h"
#endif /* NCSA */
#else /* DOS */
#include <sys/types.h>
#include <sys/time.h>
#endif /* DOS */
#endif /* MACOS */

#ifdef MACOS
#define tcp_close( s )		tcpclose( s )
#else /* MACOS */
#ifdef DOS
#ifdef PCNFS
#define tcp_close( s )		close( s )
#endif /* PCNFS */
#ifdef NCSA
#define tcp_close( s )		netclose( s ); netshut()
#endif /* NCSA */
#ifdef WINSOCK
#define tcp_close( s )		closesocket( s ); WSACleanup();
#endif /* WINSOCK */
#else /* DOS */
#define tcp_close( s )		close( s )
#endif /* DOS */
#endif /* MACOS */


/*
 * ldap_bind - bind to the ldap server (and X.500).  The dn and password
 * of the entry to which to bind are supplied.  0 is returned on success,
 * -1 if the DSA could not be contacted.
 *
 * Example:
 *	ldap_bind( ld, "cn=manager, o=university of michigan, c=us", "secret",
 *	    LDAP_AUTH_SIMPLE )
 */

/* ARGSUSED */
ldap_bind( ld, who, passwd, authmethod )
LDAP	*ld;
char	*who;
char	*passwd;
int	authmethod;
{
	BerElement	*ber;
	LDAPMessage	*result;
	char		*cred;
	int		credlen;
#ifdef KERBEROS
	char		*get_kerberosv4_credentials();
#endif /* KERBEROS */

	/*
	 * The bind request looks like this:
	 *	BindRequest ::= SEQUENCE {
	 *		version		INTEGER,
	 *		name		DistinguishedName,	 -- who
	 *		authentication	CHOICE {
	 *			simple		[0] OCTET STRING -- passwd
#ifdef KERBEROS
	 *			krbv42ldap	[1] OCTET STRING
	 *			krbv42dsa	[2] OCTET STRING
#endif
	 *		}
	 *	}
	 * all wrapped up in an LDAPMessage sequence.
	 */

	Debug( LDAP_DEBUG_TRACE, "ldap_bind\n", 0, 0, 0 );

	if ( who == NULL )
		who = "";
	if ( passwd == NULL )
		passwd = "";

	switch ( authmethod ) {
	case LDAP_AUTH_SIMPLE:
		cred = passwd;
		credlen = strlen( passwd );
		break;

#ifdef KERBEROS
	case LDAP_AUTH_KRBV4:
		cred = get_kerberosv4_credentials( ld, who, passwd,
		    "ldapserver", &credlen );
		break;
#endif

	default:
		ld->ld_errno = LDAP_AUTH_UNKNOWN;
		return( -1 );
		break;
	}
	if ( cred == NULL ) { /* ld_errno set in get_kerberosv4_credentials */
		return( -1 );
	}

	/* create a message to send */
	if ( (ber = ber_alloc()) == NULLBER ) {
		ld->ld_errno = LDAP_LOCAL_ERROR;
		return( -1 );
	}

	/* fill it in */
	if ( ber_printf( ber, "{it{isto}}", ++ld->ld_msgid, LDAP_REQ_BIND,
	    ld->ld_version, who, authmethod, cred, credlen ) == -1 ) {
		ld->ld_errno = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( -1 );
	}

	/* send the message */
	if ( ber_flush( &ld->ld_sb, ber, 1 ) == -1 ) {
		ld->ld_errno = LDAP_SERVER_DOWN;
		return( -1 );
	}

	if ( ld->ld_cache != NULL ) {
		ldap_flush_cache( ld );
	}

	/* wait for a result */
	if ( ldap_result( ld, ld->ld_msgid, 1, (struct timeval *) 0, &result )
	    == -1 )
		return( -1 );	/* ldap_result should have set ld_errno */

	if ( (ld->ld_errno = ldap_result2error( ld, result, 1 ))
	    != LDAP_SUCCESS )
		return( ld->ld_errno );

	switch ( authmethod ) {
	case LDAP_AUTH_SIMPLE:
		return( ld->ld_errno );
		break;

#ifdef KERBEROS
	case LDAP_AUTH_KRBV4:
		cred = get_kerberosv4_credentials( ld, who, passwd,
		    "x500dsa", &credlen );
		authmethod = LDAP_AUTH_KRBV42;
		break;
#endif

	default:
		ld->ld_errno = LDAP_AUTH_UNKNOWN;
		return( -1 );
		break;
	}
	if ( cred == NULL ) { /* ld_errno set in get_kerberosv4_credentials */
		return( -1 );
	}

	/* create a message to send */
	if ( (ber = ber_alloc()) == NULLBER ) {
		ld->ld_errno = LDAP_LOCAL_ERROR;
		return( -1 );
	}

	/* fill it in */
	if ( ber_printf( ber, "{it{isto}}", ++ld->ld_msgid, LDAP_REQ_BIND,
	    ld->ld_version, who, authmethod, cred, credlen ) == -1 ) {
		ld->ld_errno = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( -1 );
	}

	/* send the message */
	if ( ber_flush( &ld->ld_sb, ber, 1 ) == -1 ) {
		ld->ld_errno = LDAP_SERVER_DOWN;
		return( -1 );
	}

	/* wait for a result */
	if ( ldap_result( ld, ld->ld_msgid, 1, (struct timeval *) 0, &result )
	    == -1 )
		return( -1 );

	return( ld->ld_errno = ldap_result2error( ld, result, 1 ) );
}

ldap_unbind( ld )
LDAP	*ld;
{
	BerElement	*ber;
	LDAPMessage	*lm, *next;
	int		err = 0;

	Debug( LDAP_DEBUG_TRACE, "ldap_unbind\n", 0, 0, 0 );

	/* create a message to send */
	if ( (ber = ber_alloc()) == NULLBER ) {
		ld->ld_errno = LDAP_LOCAL_ERROR;
		err = -1;
	}

	/* fill it in */
	if ( !err && ber_printf( ber, "{itn}", ++ld->ld_msgid, LDAP_REQ_UNBIND )
	    == -1 ) {
		ld->ld_errno = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		err = -1;
	}

	/* send the message */
	if ( !err && ber_flush( &ld->ld_sb, ber, 1 ) == -1 ) {
		ld->ld_errno = LDAP_SERVER_DOWN;
		ber_free( ber, 1 );
		err = -1;
	}

	/* free LDAP structure and outstanding requests/responses */
	for ( lm = ld->ld_requests; lm != NULL; lm = next ) {
		next = lm->lm_next;
		ldap_msgfree( lm );
	}

	for ( lm = ld->ld_responses; lm != NULL; lm = lm->lm_next ) {
		next = lm->lm_next;
		ldap_msgfree( lm );
	}

	tcp_close( ld->ld_sb.sb_sd );

	if ( ld->ld_cache != NULL )
		ldap_destroy_cache( ld );
	if ( ld->ld_error != NULL )
		free( ld->ld_error );
	if ( ld->ld_host != NULL )
		free( ld->ld_host );
	if ( ld->ld_sb.sb_ber.ber_buf != NULL )
		free( ld->ld_sb.sb_ber.ber_buf );
	free( (char *) ld );

	return( err );
}
